"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const mongoose_1 = require("mongoose");
class MongooseCounter {
    constructor(mongoose, schema, options) {
        this.options = {
            id: null,
            incField: '_id',
            collectionName: 'counters',
        };
        this.useReference = false;
        this.schema = schema;
        const counterOptions = Object.assign({}, this.options, options);
        if (counterOptions.referenceFields === undefined) {
            counterOptions.referenceFields = [counterOptions.incField];
        }
        else {
            this.useReference = true;
        }
        counterOptions.referenceFields = Array.isArray(counterOptions.referenceFields)
            ? counterOptions.referenceFields
            : [counterOptions.referenceFields];
        if (this.useReference === true && counterOptions.id === null) {
            throw new Error('Cannot use reference fields without specifying a name');
        }
        else {
            counterOptions.id = counterOptions.id || counterOptions.incField;
        }
        this.options = counterOptions;
        this.counterModel = this.createCounterModel(mongoose);
    }
    initialize() {
        this.addCounterToSchema();
        this.addStaticMethods();
        this.addHooks();
    }
    /**
     * Create a model for the counter.
     *
     * @private
     * @param {Mongoose} mongoose
     * @returns
     * @memberof MongooseCounter
     */
    createCounterModel(mongoose) {
        const { collectionName, id } = this.options;
        const modelName = `${collectionName.charAt(0).toUpperCase()}${collectionName.slice(1)}_${id}`;
        const CounterSchema = new mongoose.Schema({
            id: {
                type: String,
                required: true,
            },
            reference: {
                type: mongoose_1.Schema.Types.Mixed,
                required: true,
            },
            counter: {
                type: Number,
                default: 0,
                required: true,
            },
        }, {
            collection: collectionName,
            validateBeforeSave: false,
            versionKey: false,
            _id: false,
        });
        // @Compound indexes
        CounterSchema.index({
            id: 1,
            reference: 1,
        }, {
            unique: true,
            sparse: true,
        });
        return mongoose.model(modelName, CounterSchema);
    }
    /**
     * Enrich the schema with keys needed by this counter.
     *
     * @private
     * @memberof MongooseCounter
     */
    addCounterToSchema() {
        const { incField } = this.options;
        const schemaKey = this.schema.path(incField);
        if (schemaKey === undefined) {
            this.schema.add({
                [incField]: Number,
            });
        }
        else if (schemaKey.instance !== 'Number') {
            throw new Error(`Auto increment field "${incField}" already present and not of type "Number"`);
        }
    }
    /**
     * Set some useful methods on the schema.
     *
     * @private
     * @memberof MongooseCounter
     */
    addStaticMethods() {
        const parent = this;
        this.schema.static('resetCounter', function (id, reference, cb) {
            const condition = { id };
            const next = reference instanceof Function ? reference : cb;
            if (!(reference instanceof Function)) {
                condition.reference = parent.getReferenceField(reference);
            }
            return parent.counterModel
                .where(condition)
                .setOptions({ multi: true })
                .update({ $set: { counter: 0 } }, next);
        });
    }
    /**
     * Set and handler for hooks on the schema referenced by this sequence.
     *
     * @private
     * @memberof MongooseCounter
     */
    addHooks() {
        const parent = this;
        const { id, incField } = this.options;
        this.schema.pre('save', async function (next) {
            if (!this.isNew)
                return next();
            try {
                const reference = parent.getReferenceField(this.toJSON());
                const res = await parent.counterModel.findOneAndUpdate({ id, reference }, { $inc: { counter: 1 } }, { new: true, upsert: true });
                this.set(incField, res.counter);
                next();
            }
            catch (err) {
                next(err);
            }
        });
    }
    /**
     * Given a mongoose document values of the fields set as reference for the counter.
     *
     * @private
     * @param StringMap document A mongoose document values
     * @returns An array of strings which represent the value of the reference
     * @memberof MongooseCounter
     */
    getReferenceField(document) {
        const { referenceFields } = this.options;
        if (!this.useReference || referenceFields === undefined)
            return null;
        return referenceFields.reduce((map, f) => {
            map[f] = document[f];
            return map;
        }, {});
    }
}
/**
 * @param {Schema} schema Schema object passed by Mongoose Schema.plugin
 * @param {*} [opts] Options passed by Mongoose Schema.plugin
 * @param {Mongoose} [options.mongoose] Mongoose instance to use
 * @param {string|string[]} [opts.omit] fields to omit from diffs (ex. ['a', 'b.c.d'])
 */
function mongooseCounter(mongoose) {
    if (!(mongoose instanceof mongoose_1.Mongoose)) {
        throw new Error('Please, pass mongoose while requiring mongoose-counters');
    }
    return function (schema, options) {
        const counter = new MongooseCounter(mongoose, schema, options);
        counter.initialize();
        return counter;
    };
}
exports.default = mongooseCounter;
